import dotmap
import yaml
import logging

from . import exceptions

log = logging.getLogger('root')


def load_YAML(file_object):
    try:
        config = yaml.safe_load(file_object)
    except yaml.composer.ComposerError as e:
        log.error(f"YAML composer error {e}")
        raise exceptions.ConfigurationError(e) from e
    except yaml.scanner.ScannerError as e:
        log.error(f"YAML scanner error {e}")
        raise exceptions.ConfigurationError(e) from e

    return dotmap.DotMap(config, _dynamic=True)


def make_static(dm):
    for k in dm._map:
        if isinstance(dm[k], dotmap.DotMap):
            make_static(dm[k])
        if isinstance(dm[k], list):
            for item in dm[k]:
                make_static(item)
        dm._dynamic = False



import abc
import argparse
import dotmap
import logging
import sys

import schema

from . import scala
from . import exceptions
from . import configuration
from . import colour as c


log = logging.getLogger("root")


class Scalyca(scala.Scala, metaclass=abc.ABCMeta):
    """
        Simple Console Application with Logging, Yaml Configuration and Argparse
    """
    _app_name = 'Scalyca'
    _description = "Simple Configurable Application with Logging, Yaml file Configuration and Argparse"
    # Config validation schema, or None to skip validation
    _prog = "Scalyca"
    schema = None

    def _add_default_arguments(self):
        """ Scalyca adds one mandatory argument: the path to the config file """
        super()._add_default_arguments()
        self.add_argument('config', type=argparse.FileType('r'), help="main configuration file")

    def _configure(self):
        try:
            self._load_configuration()
            self._override_configuration()
            self.override_configuration()
            self._lock_configuration()
            self._debug_configuration()
            self._validate_configuration()

        except exceptions.FatalError:
            log.critical(f"{c.script(self._app_name)} aborting during configuration")
            sys.exit(-1)

    def _load_configuration(self):
        raw = configuration.load_YAML(self.args.config)
        self.config = dotmap.DotMap(raw, _dynamic=True)
        log.debug(f"Configuration read from {c.path(self.args.config.name)}")

    def _lock_configuration(self):
        log.debug(f"Locking the configuration")
        configuration.make_static(self.config)

    def override_with_warning(self, parameter, new_value, name):
        log.warning(f"Overriding {c.param(name)} ({c.over(parameter)} -> {c.over(new_value)})")
        self.config[name] = new_value

    def _debug_configuration(self):
        if self.args.debug:
            log.debug("Full config is")
            self.config.pprint()

    def _validate_configuration(self):
        if self.schema is None:
            log.debug("No config validation schema provided, skipping validation")
        else:
            try:
                self.schema.validate(self.config)
            except schema.SchemaError as e:
                raise exceptions.ConfigurationError(e) from e



import abc
import argparse
import dotmap
import logging
import yaml
import sys

from . import colour as c
from . import exceptions
from . import logger
from . import configuration

log = logger.setup_log('root', timefmt='%H:%M:%S')


class Scala(metaclass=abc.ABCMeta):
    """
        Scala: Simple Console Application with Logging and Argparse
    """
    _app_name = "Default Scala"
    _description = "Simple Configurable Application with Logging and Argparse"
    _prog = "Scala"

    def __init__(self, **kwargs):
        """ Optionally override the application name and description """
        self._ok = True
        self._description = kwargs.get('description', self._description)
        self._app_name = kwargs.get('app_name', self._app_name)
        self._prog = kwargs.get('app_name', self._prog)
        self._success_message = kwargs.get('success_message', f"{c.script(self._app_name)} finished successfully")

        self._argparser = None

    def _initialize(self):
        self._argparser = argparse.ArgumentParser(
            prog=self._prog,
            description=self._description
        )
        self._add_default_arguments()
        self.add_arguments()
        self.args = self._argparser.parse_args()

        self._configure()

    def add_argument(self, *args, **kwargs):
        """ Just a public wrapper so that we do not have to access _argparser directly """
        self._argparser.add_argument(*args, **kwargs)

    @abc.abstractmethod
    def add_arguments(self):
        """ Override this in your Scala derivative to add custom arguments """

    def _add_default_arguments(self):
        """ Add default arguments that are present in every instance of Scala """
        self.add_argument('-l', '--logfile', type=argparse.FileType('w'), help="Write log to file")
        self.add_argument('-d', '--debug', action='store_true', help="Turn on verbose logging")

    def initialize(self):
        """ Custom initialization. Might be empty. """

    @abc.abstractmethod
    def main(self):
        """
        The main entry point of the program.
        Provide an implementation in your derivative of Scala.
        """

    def _configure(self):
        pass

    def _override_configuration(self):
        log.setLevel(logging.DEBUG if self.args.debug else logging.INFO)

        if self.args.logfile:
            log.addHandler(logging.FileHandler(self.args.logfile.name))
            log.debug(f"Added log output {c.path(self.args.logfile.name)}")

    def run(self):
        try:
            self._initialize()
            self.initialize()
            self.main()
            self._finalize()
        except exceptions.PrerequisiteError as e:
            log.error(f"Terminating due to missing prerequisites: {e}")
        except exceptions.ConfigurationError as e:
            log.error(f"Terminating due to a configuration error: {e}")
        finally:
            if self._ok:
                log.info(self._success_message)
            else:
                log.critical(f"{c.script(self._app_name)} aborted during runtime")

    def _finalize(self):
        self._ok = True
